#region usings
using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Drawing;
using System.IO;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;

using SlimDX;
using SlimDX.Direct3D9;
using VVVV.Core.Logging;
using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V2.EX9;
using VVVV.Utils.SlimDX;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;

#endregion usings

namespace VVVV.Nodes
{
	//little helper class used to store information for each
	//texture resource
	public class FrameData
	{
		public int Width;
		public int Height;
		public byte[] Pixels;
		public bool Copied;
	}
	
	
	#region PluginInfo
	[PluginInfo(Name = "ImageLoader", Category = "EX9.Texture", Help = "Loads image files into textures using the Windows.Media API", Tags = "bmp, gif, icon, png, tiff, wmp")]
	#endregion PluginInfo
	public class EX9_TextureImageLoader : IPluginEvaluate, IPartImportsSatisfiedNotification
	{
		[Input("Filename", DefaultValue = 1, StringType = StringType.Filename)]
		public IDiffSpread<string> FFilenameIn;
		
		[Input("Reload", IsBang = true)]
		public IDiffSpread<bool> FReloadIn;
		
		[Output("Texture Out", BinName = "Frame Count")]
		public ISpread<ISpread<TextureResource<FrameData>>> FTextureOut;
		
		[Output("Original Format")]
		public ISpread<string> FFormatOut;
		
		[Output("Loading")]
		public ISpread<bool> FLoadingOut;
		
		[Output("Progress")]
		public ISpread<double> FProgressOut;
		
		[Import()]
		public ILogger FLogger;
		private Spread<ImageLoaderSlice> FGifDecoders = new Spread<ImageLoaderSlice>(0);
		
		public void OnImportsSatisfied()
		{
			//spreads have a length of one by default, change it
			//to zero so ResizeAndDispose works properly.
			FTextureOut.SliceCount = 0;
		}
		
		//called when data for any output pin is requested
		public void Evaluate(int spreadMax)
		{
			//resize and create
			FGifDecoders.ResizeAndDismiss(spreadMax, () => { return new ImageLoaderSlice(); });
			
			if(FFilenameIn.IsChanged || FReloadIn.IsChanged)
			{
				FTextureOut.SliceCount = spreadMax;
				FLoadingOut.SliceCount = spreadMax;
				FFormatOut.SliceCount = spreadMax;
				FProgressOut.SliceCount = spreadMax;
				
				for (int i = 0; i < spreadMax; i++)
				{
					var decoder = FGifDecoders[i];
					
					//set params
					decoder.OpenFileAsync(FFilenameIn[i], FReloadIn[i]);
					
					//set outputs
					FTextureOut[i] = decoder.TextureSpread;
					FFormatOut[i] = decoder.Format.ToString();
				}
			}
			
			//set loading
			for (int i = 0; i < spreadMax; i++)
			{
				var decoder = FGifDecoders[i];
				FLoadingOut[i] = decoder.IsLoading;
				FProgressOut[i] = decoder.Progress;
			}
			
		}
	}
	
	public class ImageLoaderSlice
	{	
		
		//properties
		private Spread<TextureResource<FrameData>> FTextureSpread = new Spread<TextureResource<FrameData>>(0);
		public ISpread<TextureResource<FrameData>> TextureSpread
		{
			get
			{
				return FTextureSpread;
			}
		}
		
		public PixelFormat Format
		{
			get;
			private set;
		}
		
		public bool IsLoading
		{
			get;
			private set;
		}
		
		public double Progress
		{
			get;
			private set;
		}

		//constructor
		bool FFirstLoad;
		public ImageLoaderSlice()
		{
			FFirstLoad = true;
		}

		List<FrameData> FFrameDataList;
		string FLastFilename = "";
		public async void OpenFileAsync(string filename, bool forceReload = false)
		{
			//valid file name?
			if(string.IsNullOrEmpty(filename) || (FLastFilename == filename && !forceReload)) return;
			
			IsLoading = true;
			
			//dispose all textures
			FTextureSpread.ResizeAndDispose(0, i => CreateTextureResource(i));
			
			FLastFilename = filename;
			
			//start async loading
			FFrameDataList = await Task.Run( () => CreateFrameData(filename, forceReload));
			
			//after loading continue here
			
			//recreate texture resources
			FTextureSpread.ResizeAndDispose(FFrameDataList.Count, i => CreateTextureResource(i));
			IsLoading = false;
		}
		
		//async loading task
		private List<FrameData> CreateFrameData(string filename, bool forceReload = false)
		{
			Progress = 0;
			var creationOptions = (forceReload || FFirstLoad) ? BitmapCreateOptions.IgnoreImageCache : BitmapCreateOptions.None;
			FFirstLoad = false;
			
			//create decoder
			BitmapDecoder decoder;
			try
			{
				decoder = BitmapDecoder.Create(new Uri(filename, UriKind.RelativeOrAbsolute), creationOptions, BitmapCacheOption.Default);
			}
			catch
			{
				return new List<FrameData>(0);
			}
			
			if(decoder.IsDownloading)
			{	
				//wait for download
				var signal = new SemaphoreSlim(0, 1);
				
				decoder.DownloadCompleted += delegate 
				{ 
					signal.Release(); 
				};
				decoder.DownloadProgress += decoder_DownloadProgress;
				decoder.DownloadFailed += delegate 
				{ 
					signal.Release(); 
				};
				
				//wait for loading complete
				signal.Wait();
			}
			
			//continue if loaded
			Progress = 1;
			var list = new List<FrameData>(decoder.Frames.Count);
			
			Format = decoder.Frames[0].Format;
			
			//convert and copy frames
			foreach (var frame in decoder.Frames) 
			{
				var convertedBitmap = new FormatConvertedBitmap();
				
				convertedBitmap.BeginInit();
				convertedBitmap.Source = frame;
				convertedBitmap.DestinationFormat = PixelFormats.Bgra32;
				convertedBitmap.EndInit();
				
				list.Add(CopyBitmapFrameToFrameData(convertedBitmap));
			}
			
			return list;
		}
		
		//set download progress
		void decoder_DownloadProgress(object sender, DownloadProgressEventArgs e)
		{
			Progress = e.Progress/100.0;
		}
		
		//create textures
		TextureResource<FrameData> CreateTextureResource(int slice)
		{	
			var res = TextureResource.Create(FFrameDataList[slice], CreateTexture, UpdateTexture);
			res.NeedsUpdate = false;
			return res;
		}
		
		//this method gets called, when Reinitialize() was called in evaluate,
		//or a graphics device asks for its data
		Texture CreateTexture(FrameData info, Device device)
		{
			info.Copied = false;
			return TextureUtils.CreateTexture(device, Math.Max(info.Width, 1), Math.Max(info.Height, 1));
		}
		
		
		//this method gets called, when Update() was called in evaluate,
		//or a graphics device asks for its texture, here you fill the texture with the actual data
		//this is called for each renderer, careful here with multiscreen setups, in that case
		//calculate the pixels in evaluate and just copy the data to the device texture here
		unsafe void UpdateTexture(FrameData info, Texture texture)
		{
			if(!info.Copied)
			{
				CopyByteArrayToTexture(info.Pixels, texture);
				info.Copied = true;
			}
			
		}
		
		//copy helpers
		private FrameData CopyBitmapFrameToFrameData(BitmapSource bm)
		{

			int stride = bm.PixelWidth * (bm.Format.BitsPerPixel / 8);
			var bytes = new byte[stride * bm.PixelHeight];
			try
			{
				bm.CopyPixels(bytes, stride, 0);
			}
			catch(Exception e)
			{
				throw e;
			}
			
			return new FrameData{ Width = bm.PixelWidth, Height = bm.PixelHeight, Pixels = bytes };
		}
		
		public static void CopyByteArrayToTexture(byte[] bm, Texture texture)
		{
			var rect = texture.LockRectangle(0, LockFlags.None);
			try
			{
				Marshal.Copy(bm, 0, rect.Data.DataPointer, (int)rect.Data.Length);
			}
			catch(Exception e)
			{
				throw e;
			}
			finally
			{
				texture.UnlockRectangle(0);
			}
			
		}
		
	}
}
